'''
Created on Apr 26, 2019

@author: baumannt
'''

from drever.base_class import Drever
from drever.data_handlers.image import ImageData


class VideoFilter(Drever):
    '''
    classdocs
    '''

    CLASS_CONFIG = {
        "INPUT_VECTOR_TYPE":  (ImageData, ),
        "OUTPUT_VECTOR_TYPE": (ImageData, )
    }

    # Overwrite this dictionary with your default Drever algorithm parameters.
    DEFAULT_PARAMS = {}

    def run(self):
        self._set_output_data(self.get_input_data())

    @classmethod
    def dump_vectors(cls, path, vectors):
        vectors.save(path, False)



'''
Created on Apr 26, 2019

@author: baumannt
'''
from drever.base_class import Drever


class Bypass(Drever):
    '''
    classdocs
    '''

    CLASS_CONFIG = {
        "INPUT_VECTOR_TYPE":  "ALL",
        "OUTPUT_VECTOR_TYPE": "ALL"
    }

    def run(self):
        '''
        The Bypass Drever simply sets the output data as the Input Data
        '''
        self._set_output_data(self.get_input_data())



'''
Created on Apr 16, 2019

@author: baumannt
'''

from abc import ABC, abstractmethod
import os


class Drever(ABC):
    '''
    This is the drever base class. New Drever classes needs to be derived from
    this base class with an overwritten run method.

    When deriving your new Drever class, you have to set configuration
    parameters by using the CLASS_CONFIG dictionary. The "INPUT_VECTOR_TYPE"
    and the "OUTPUT_VECTOR_TYPE" needs to be defined. Valid values are classes,
    e.g. a Numpy array using "numpy.ndarray" or tuples of classes. If you don't
    care about the IN/OUT data types you can use the "ALL" string, which
    disables type checking (for an example see the Bypass class).
    '''

    # Overwrite this dictionary in your derived Drever class.
    CLASS_CONFIG = {
        "INPUT_VECTOR_TYPE":  None,
        "OUTPUT_VECTOR_TYPE": None
    }

    # Overwrite this dictionary with your default Drever algorithm parameters.
    DEFAULT_PARAMS = {}

    def __init__(self, params=None, input_data=None, dump_filename=None):
        '''
        Constructor
        '''
        self.params = self.DEFAULT_PARAMS if params is None else params
        self.dump_filename = None if dump_filename is None else dump_filename

        # Never use this variables without the getters and setters.
        self._input_data = None
        self._output_data = None

        if input_data is not None:
            self.set_input_data(input_data)

    @abstractmethod
    def run(self):
        '''
        This is the main method of your Drever instance. Overwrite the run
        method with your algorithm.
        '''

    @classmethod
    def dump_vectors(cls, path, vectors):
        '''
        This method stores the vector data. The dump can be used inside the
        HDL testbench for verification.

        Overwrite this method for proper vector dumping.
        '''
        assert vectors is not None, "Vector contains no data!"

        assert False, \
            "Vector dumpin not overloaded. No vector file was written to\n" \
            "Path:    " + path + "\n\n"

    def dump_input_vectors(self, path):
        self.dump_vectors(path, self.get_input_data())

    def dump_output_vectors(self, path):
        self.dump_vectors(path, self.get_output_data())

    def set_params(self, params):
        '''
        Parameter Setter
        '''
        self.params = params

    def get_params(self):
        '''
        Parameter Getter
        '''
        return self.params

    def set_input_data(self, data):

        if self.CLASS_CONFIG["INPUT_VECTOR_TYPE"] != "ALL":
            assert self.CLASS_CONFIG["INPUT_VECTOR_TYPE"] is not None, \
                "INPUT_VECTOR_TYPE of your derived Drever isn't set!"
            assert isinstance(data, self.CLASS_CONFIG["INPUT_VECTOR_TYPE"]), \
                "Input data can't be set, due invalid input data type!"

        self._input_data = data

    def get_input_data(self):

        if self.CLASS_CONFIG["INPUT_VECTOR_TYPE"] != "ALL":
            assert isinstance(
                self._input_data,
                self.CLASS_CONFIG["INPUT_VECTOR_TYPE"]), \
                "Data can't be read, due the class' " \
                "input data has invalid type!"
        return self._input_data

    def _set_output_data(self, data):

        if self.CLASS_CONFIG["OUTPUT_VECTOR_TYPE"] != "ALL":
            assert self.CLASS_CONFIG["OUTPUT_VECTOR_TYPE"] is not None, \
                "OUTPUT_VECTOR_TYPE of your derived Drever isn't set!"
            assert isinstance(data, self.CLASS_CONFIG["OUTPUT_VECTOR_TYPE"]), \
                "Output data can't be set, due invalid input data type!"

        self._output_data = data

    def get_output_data(self):

        if self.CLASS_CONFIG["OUTPUT_VECTOR_TYPE"] != "ALL":
            assert isinstance(
                self._output_data,
                self.CLASS_CONFIG["OUTPUT_VECTOR_TYPE"]), \
                "Data can't be read, due the class' output data has " \
                "invalid type! Maybe run() method wasn't called yet?"

        return self._output_data

    def set_dump_filename(self, dump_filename):
        assert dump_filename, "Non valid filename string"
        self.dump_filename = dump_filename

    def vunit_get_dump_filename(self, data=None):

        assert data in ("input", "output", None), \
            "Data source for dump file not know!"

        default = "output"
        result = ""

        if data == "input":
            result = data + "_" + self.dump_filename
        else:
            result = default + "_" + self.dump_filename

        return result

    # This is covered within the HDL tests
    def vunit_pre_config(self, output_path):  # pragma: no cover
        stimulus_filename = os.path.join(
            output_path,
            self.vunit_get_dump_filename("input")
        )
        processed_filename = os.path.join(
            output_path,
            self.vunit_get_dump_filename("output")
        )

        self.dump_input_vectors(stimulus_filename)
        self.run()
        self.dump_output_vectors(processed_filename)

        return True

    # This is covered within the HDL tests
    def vunit_post_check(self, output_path):  # pragma: no cover
        return False

    def vunit_generate_generics(self):
        return dict(
            IN_FILENAME=self.vunit_get_dump_filename("input"),
            OUT_FILENAME=self.vunit_get_dump_filename("output")
        )



'''
Created on Apr 4, 2019

Package Information Module

@author: baumannt
'''

MAJOR_VERSION = 0
MINOR_VERSION = 0
PATCH_VERSION = 3


def get_version():
    '''
    Getter for version string of package.

    Returns:
    str:Version Number
    '''

    version_str = '.'.join(
        [str(MAJOR_VERSION), str(MINOR_VERSION), str(PATCH_VERSION)])
    return version_str


def print_version():

    ''' Prints version on stdout '''

    print("Version: " + get_version())



'''
Created on Apr 12, 2019

Python __init__

@author: baumannt
'''
import os
from vunit import VUnit

from drever import pkg_info

__version__ = pkg_info.get_version()

VHDL_LIBRARY_NAME = "drever"
VHDL_SRC_DIR = os.path.join(__path__[0], "vhdl/**/*.vhd")


def add_drever_hdl_sources(vunit):  # pragma: no cover
    drever_lib = vunit.add_library(VHDL_LIBRARY_NAME)
    drever_lib.add_source_files(VHDL_SRC_DIR, vhdl_standard="2008")



import setuptools
import drever

vhdl_files = [
    "vhdl/*.vhd",
    "vhdl/**/*.vhd"
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="drever",
    version=drever.__version__,
    author="Tobias Baumann",
    author_email="tobias.baumann@elpra.de",
    description="A framework for handling testvectors " +
                "within an HDL testbench.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/Elpra/drever-framework/drever",
    packages=setuptools.find_packages(),
    package_data={"drever": vhdl_files},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License " +
        "v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
)



'''
2D Image Data
'''
import os
import errno
import math
import numpy as np


MIN_IMAGE_WIDTH = 1
MIN_IMAGE_HEIGHT = 1
# Set maxmimum image size to 4K resolution
MAX_IMAGE_WIDTH = 3840
MAX_IMAGE_HEIGHT = 2160
MAX_BITDEPTH = 16


class ImageData():
    '''
    classdocs
    '''

    # Define valid ranges for the several ImageData parameters
    PARAMS_RANGE = {
        "width":    range(MIN_IMAGE_WIDTH, MAX_IMAGE_WIDTH+1),
        "height":   range(MIN_IMAGE_HEIGHT, MAX_IMAGE_HEIGHT+1),
        "bitdepth": range(1, 17),
        "channels": range(1, 4)  # e.g. 1 for gray image,
                                 # 2 for 422 YUV, 3 for RGB image
    }

    COMMON_BITDEPTHS = [1, 8, 10, 12]

    def __init__(self, params=None):
        '''
        Constructor
        '''
        self.params = {
            "width":    0,
            "height":   0,
            "bitdepth": 8,
            "channels": 3
        }
        self.image_data = None
        self.set_params(params)

    def __getitem__(self, key):
        return self.get_image_data()[key]

    def __setitem__(self, key, value):
        self.get_image_data()[key] = value

    def __eq__(self, other):

        objs = (
            {"obj": self, "data": None, "image": False},
            {"obj": other, "data": None, "image": False}
        )

        for obj in objs:

            if isinstance(obj["obj"], ImageData):
                obj["data"] = obj["obj"].get_image_data()
                obj["image"] = True
            elif isinstance(obj["obj"], np.ndarray):
                obj["data"] = obj["obj"]
            else:
                return NotImplemented

        if not np.array_equal(objs[0]["data"], objs[1]["data"]):
            return False

        if (objs[0]["image"] and objs[1]["image"]) is True:
            if self.params != other.params:
                return False

        return True

    def check_params(self, params, use_assertions=False):

        keys_are_equal = set(params.keys()) == set(self.PARAMS_RANGE.keys())

        if use_assertions:
            assert keys_are_equal, \
                "Parameter keys are invalid!"

        if not keys_are_equal:
            return False

        if use_assertions:
            for key, value in params.items():
                assert value in self.PARAMS_RANGE[key], \
                    'Parameter "' + key + '" is out of range!'

        for key, value in params.items():
            if value not in self.PARAMS_RANGE[key]:
                return False

        return True

    def set_params(self, params):

        if params is None:
            self.params = None
        elif self.check_params(params, True):
            self.params = params

    @classmethod
    def extract_params(cls, data, use_bitdepth=None):

        assert data.dtype == np.uint16, \
            "Data is not np.uint16"

        assert len(data.shape) in range(2, 4), \
            "Data has unsupported array dimensions!"

        max_value = np.amax(data)

        if use_bitdepth == "Real":
            bitdepth = math.ceil(math.log2(max_value))
        elif use_bitdepth in cls.PARAMS_RANGE["bitdepth"]:

            bitdepth = use_bitdepth

            assert max_value < 2**bitdepth, \
                "Given bitdepth missmatches the maximum image_data value"

        else:

            depths = cls.COMMON_BITDEPTHS

            # If no common bitdepth can be found, use the maximum bitdepth
            bitdepth = MAX_BITDEPTH

            for test_bitdepth in depths:
                if max_value < 2**test_bitdepth:
                    bitdepth = test_bitdepth
                    break

        if len(data.shape) == 2:
            num_channels = 1
        else:

            assert data.shape[2] in range(1, 4), \
                "Data has unsupported array dimensions!"

            num_channels = data.shape[2]

        return {
            "width":    data.shape[1],
            "height":   data.shape[0],
            "bitdepth": bitdepth,
            "channels": num_channels
        }

    def get_np_shape(self):
        return (
            self.params['height'],
            self.params['width'],
            self.params['channels']
        )

    def check_data(self, data, use_bitdepth=None):

        assert data.dtype == np.uint16, \
            "Data is not np.uint16"

        extracted_params = self.extract_params(data, use_bitdepth)

        if extracted_params != self.params:
            print("*************************************************")
            print("Data Check Error")
            print("*************************************************")
            print("- Extracted Parameters:")
            print(extracted_params)
            print("- Set Parameters:")
            print(self.params)
            print("*************************************************")
            return False

        return True

    def init_with_data(self, data, extract_params=True, use_bitdepth=None):

        if extract_params:
            self.params = self.extract_params(data, use_bitdepth)
        else:
            assert self.check_data(data, use_bitdepth), \
                "Check Data failed!"

        assert self.params is not None, \
            "Image parameters needs to be set first!"

        self.image_data = data.reshape(
            (self.params['height'],
             self.params['width'],
             self.params['channels'])
        )

    def init_with_random(self, seed=None):

        assert self.params is not None, \
            "Image parameters needs to be set first!"

        upper_limit = 2**self.params['bitdepth']

        if seed is not None:
            np.random.seed(seed)

        dim = self.get_np_shape()

        rnd_data = (upper_limit*np.random.rand(*dim)).astype(np.uint16)

        self.init_with_data(rnd_data, False, None)

    def get_image_data(self):
        return self.image_data

    def load(self, path):
        '''
        THIS IS JUST A DUMMY AND NOT IMPLEMENTED YET.

        Loads image data (numpy array) from ppm file.

        Parameters:
        path (str): Path of ppm file
        '''

        if not os.path.isfile(path):
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), path)

        ppm_file = open(path, "r")

        states = [
            "HEADER_MAGIC_NUMBER",
            "HEADER_GEOMETRY",
            "HEADER_MAX_VALUE",
            "DATA"
        ]

        image_params = {
            "width":    0,
            "height":   0,
            "bitdepth": 8,
            "channels": 3
        }
        image_data = None
        data_coord = [0, 0, 0]
        chan_3_is_zero = True
        num_values = 0
        i = 0

        for raw_line in ppm_file:

            line = raw_line.strip()

            # skip line if empty
            if line == "":
                continue

            if states[0] == "HEADER_MAGIC_NUMBER":

                assert line in ("P2", "P3"), \
                    "Unsupported Magic Number: " + line

                image_params["channels"] = 1 if line == "P2" else 3
                states.remove("HEADER_MAGIC_NUMBER")

            elif states[0] == "HEADER_GEOMETRY":
                image_params["width"], image_params["height"] = \
                    [int(str_data) for str_data in line.split()]
                num_values = \
                    image_params["width"] * \
                    image_params["height"] * \
                    image_params["channels"]
                states.remove("HEADER_GEOMETRY")

            elif states[0] == "HEADER_MAX_VALUE":
                image_params["bitdepth"] = \
                    math.floor(math.log(int(line), 2)) + 1
                image_data = np.zeros(
                    (image_params["height"], image_params["width"], 3),
                    np.uint16)
                states.remove("HEADER_MAX_VALUE")

            else:
                data = [int(str_data) for str_data in line.split()]

                for value in data:

                    i += 1

                    # pylint: disable=E1137
                    image_data[
                        data_coord[0],
                        data_coord[1],
                        data_coord[2]
                    ] = value

                    if data_coord[2] == 2 and value != 0:
                        chan_3_is_zero = False

                    data_coord[2] += 1
                    if data_coord[2] == image_params["channels"]:
                        data_coord[2] = 0
                        data_coord[1] += 1
                        if data_coord[1] == image_params["width"]:
                            data_coord[1] = 0
                            data_coord[0] += 1

        if chan_3_is_zero and image_params["channels"] == 3:
            image_params["channels"] = 2

        ppm_file.close()

        assert i == num_values, \
            "Read Error: Number of pixel does not match header"

        self.set_params(image_params)
        self.init_with_data(
            image_data[::, ::, 0:image_params["channels"]],
            False,
            image_params["bitdepth"]
        )

    def save(self, path, allow_overwrite=False):
        '''
        Stores image data (numpy array) as ppm file.

        Parameters:
        path (str): Path of ppm file
        allow_overwrite (bool): When true and path already exists, file will
                                be overwritten.
        '''

        if not allow_overwrite and os.path.isfile(path):
            raise FileExistsError(
                errno.EEXIST, os.strerror(errno.EEXIST), path)

        # Set Magic Number in Header
        if self.params['channels'] == 1:
            header = "P2\n"
        else:
            header = "P3\n"

        max_value = 2**self.params['bitdepth'] - 1
        ascii_size = math.floor(math.log(max_value, 10)) + 1

        # Set Width, Height and Maximum Value in Header
        header += str(self.params['width']) + " "
        header += str(self.params['height']) + "\n"
        header += str(max_value) + "\n"

        ppm_file = open(path, "w")
        ppm_file.write(header)

        # Write Image Data
        for row in range(0, self.params['height']):

            line = ""

            for col in range(0, self.params['width']):

                value = ""

                for rgb in range(0, self.params['channels']):
                    data_str = str(self.image_data[row, col, rgb])
                    value += data_str.rjust(ascii_size, " ") + " "

                # For 2 channel images, set third channel 0
                value += "0 " if self.params['channels'] == 2 else ""
                line += value

            ppm_file.write(line + "\n")

        ppm_file.close()
